#!/usr/bin/env Rscript

## To load in interactive R shell, run: source('classifier-analysis.R')

## install.packages(c("broom", "cowplot", "dplyr", "ggplot2", "scales", "tidyr", "xtable")

## Useful web sites: http://r4ds.had.co.nz/ , https://www.tidyverse.org/
##
library('broom')
# library('cowplot')
library('dplyr')
library('ggplot2')
library('scales')
library('tidyr')
library('xtable')

test_file_names = c("amazon_review_polarity","sogou_news","yahoo_answers","yelp_review_full")

## https://stackoverflow.com/questions/5831794/opposite-of-in
`%not_in%` = function (x, table) is.na(match(x, table, nomatch=NA_integer_))  

options(dplyr.width = 100)  # Number of columns to print

raw_data = bind_rows('0'=read.csv('classifier-results.csv', header=T),
                     '1'=read.csv('shuffle-results1.csv', header=T),
                     '2'=read.csv('shuffle-results2.csv', header=T),
                     '3'=read.csv('shuffle-results3.csv', header=T),
                     '4'=read.csv('shuffle-results4.csv', header=T),
                     '5'=read.csv('shuffle-results5.csv', header=T),
                     '6'=read.csv('shuffle-results6.csv', header=T),
                     '7'=read.csv('shuffle-results7.csv', header=T),
                     .id='Run') %>% mutate(Run=factor(Run))

raw_data = raw_data %>%
    filter(complete.cases(.)) %>%              # Remove NAs
    mutate(error=1.0-prec,                     # Calculate error
           Training.file=training_file,
           training.examples=size,
           config=factor(config)) %>%          # Convert config to factor
    select(-prec,-rec,-size, -training_file)   # Remove unnecessary columns

## data_eq() finds the min error configuration in each data group
## of size fraction==frac, and then selects the examples from smaller
## data sizes with that config.
##
data_eq = function(d, frac) {
## For slice(which.min()) see
## https://stackoverflow.com/questions/30024437/applying-group-by-and-summarise-on-data-while-keeping-all-the-columns-info
## https://stackoverflow.com/questions/21308436/dplyr-filter-get-rows-with-minimum-of-variable-but-only-the-first-if-multiple
    result = d %>%
        filter(fraction==frac) %>%
        group_by(Training.file, Run, config) %>%
        summarise(av.error=mean(error)) %>%
        filter(rank(av.error, ties.method="random") == 1) %>%
        select(Training.file, Run, config) %>%
        left_join(d %>% filter(fraction <= frac));
    return(result);
}

## data_leq() finds the min error configuration within each data group
## for each size less than or equal to fraction.
##
data_leq = function(d, frac) {
    result = d %>%
        filter(fraction<=frac) %>%
        group_by(Training.file, Run, training.examples, config) %>%
        summarise(av.error=mean(error)) %>%
        filter(rank(av.error, ties.method="random") == 1) %>%
        select(Training.file, Run, training.examples, config) %>%
        left_join(d %>% filter(fraction <= frac));
    return(result);
}


#################
#
# Train and evaluate models on different subsets of data

# Training data for predictive models of error
#
train_data = bind_rows(# '==1.0'=(raw_data %>% data_eq(1.0)),
                       '==0.5'=(raw_data %>% data_eq(0.5)),
                       '==0.1'=(raw_data %>% data_eq(0.1)),
                       # '<=1.0'=(raw_data %>% group_by(Training.file, training.examples) %>% slice(which.min(error))),
                       '<=0.5'=(raw_data %>% data_leq(0.5)),
                       '<=0.1'=(raw_data %>% data_leq(0.1)),
                       .id='Pilot.data') %>%
    mutate(Pilot.data=factor(Pilot.data, levels=c('==0.1','<=0.1','==0.5','<=0.5','==1.0','<=1.0')))

# data to evaluate predictive models on, includes the min error
#
eval_data = raw_data %>%
    filter(fraction==1.0) %>%
    group_by(Training.file, training.examples, config) %>%
    summarise(av.error=mean(error)) %>%
    group_by(Training.file, training.examples) %>%
    summarise(min.av.error=min(av.error))

# Points at which to plot model predictions (assuming log x scale)
#
npoints = 25  # number of points in a fitted plot

log_seq = Vectorize(function (begin, end, n) {   # returns a list of n points logarithmically spaced from begin to end inclusive
          list(10^seq(from=log10(begin), to=log10(end), length.out=n))
     }, vectorize.args=c("begin","end")) # Vectorized version of log_seq

plot_data = raw_data %>%
    group_by(Training.file) %>%
    summarise(min.training.examples=min(training.examples), max.training.examples=max(training.examples)) %>%
    mutate(training.examples=log_seq(min.training.examples, max.training.examples, npoints)) %>%
    unnest(training.examples) %>%
    select(-min.training.examples, -max.training.examples)

##########################
#
# Model training

## define selfStart function for non-linear regression
## See Ritz Streibig (2008) Nonlinear regression with R, page 42, for information on self-starter
weightfn_selfstart = function (d) { d$training.examples }
plaw_selfStart = selfStart(~ a + b*(x**(c-0.5)),
                           function(mCall, data, LHS) {
                               xy = sortedXyData(mCall[["x"]], LHS, data);
                               d = xy %>% mutate(training.examples=x, error=y);
                               lmFit = lm(error ~ I(1/sqrt(training.examples)), data=d, weights=weightfn_selfstart(d));
                               coefs = coef(lmFit);
                               a = coefs[1]; # intercept
                               b = coefs[2]; # factor
                               value = c(a, b, 0);
                               names(value) = mCall[c("a", "b", "c")];
                               return(value);
                           },
                           c("a","b","c"));

# train_models_wf() returns a data frame containing plaw, isqrt and eplaw models for a given weight fn
#
train_models_wf = function(data, weightf) {
    # https://stackoverflow.com/questions/22713325/fitting-several-regression-models-with-dplyr
    # https://cran.r-project.org/web/packages/broom/vignettes/broom.html
    # http://stat545.com/block023_dplyr-do.html
    
    # plaw_models: error = b*training.examples^c
    plaw_models = data %>%
        group_by(Training.file, Run, Pilot.data) %>%
        do(model=glm(error ~ I(log(training.examples)), data=., family=gaussian(link='log'), weights=1e-4*weightf(.)));
    
    # isqrt models: error = a+b*1/sqrt(training.examples)
    isqrt_models = data %>%
        group_by(Training.file, Run, Pilot.data) %>%
        do(model=lm(error ~ I(1.0/sqrt(training.examples)), data=., weights=weightf(.)));
    
    # eplaw models: error = a + b*training.examples^c
    eplaw_models = data %>%
        group_by(Training.file, Run, Pilot.data) %>%
        do(model = nls(error ~ plaw_selfStart(training.examples,a,b,c),
                       data=.,
                       control=nls.control(warnOnly=TRUE,maxiter=100000,tol=1e-4,minFactor=1e-7),
                       weights=weightf(.)));
    models = bind_rows('b*n^c'=plaw_models, 'a+b*n^-1/2'=isqrt_models, 'a+b*n^c'=eplaw_models, .id='Extrapolation');
    return(models %>% ungroup %>% mutate(Extrapolation=factor(Extrapolation, levels=c('b*n^c','a+b*n^-1/2','a+b*n^c'))));
}

# train_models() returns a data frame containing plaw, isqrt and eplaw models for all the weight fns
#
train_models = function(data=train_data) {
    # Also see https://rstudio-pubs-static.s3.amazonaws.com/19698_a4c472606e3c43e4b94720506e49bb7b.html
    # weight functions, which compute commonly-used weights
    weightfn_1 = function (d) { rep(1.0, length(d$training.examples)) }
    weightfn_size = function (d) { d$training.examples }
    weightfn_ivar = function (d) { d$training.examples/(d$error*(1-d$error)) }

    models = bind_rows('1'=train_models_wf(data, weightfn_1),
                       'n'=train_models_wf(data, weightfn_size),
                       'n/e*(1-e)'=train_models_wf(data, weightfn_ivar),
                       .id='Weighting');
    return(models %>% ungroup %>% mutate(Weighting=factor(Weighting, levels=c('1','n','n/e*(1-e)'))));
}

models = train_models()

# predictions() predicts error on eval_data using models
#
# Arguments:
#
#  data is the data on which model predictions are to be made
#
#  models is a data frame, where the column `model` contains a model to
#   apply to data
#
## predictions = function(data, models) {
##     # See solution 2 in
##     # https://stackoverflow.com/questions/39836916/how-can-i-apply-grouped-data-to-grouped-models-using-broom-and-dplyr
##     preds =
##         data %>%
##         group_by(Training.file) %>%
##         nest(.key=nested_data) %>%
##         inner_join(models, .) %>%
##         mutate(preds = list(augment(model, newdata=nested_data))) %>%
##         unnest(preds)
##     return(preds %>% select(-model, -nested_data) %>% rename(predicted.error=.fitted));
## }

# This function assumes that the same model is associated with each (Training.file,Run,Pilot.data,Extrapolation,Weighting)
#
predictions = function(data, models) {
    preds = data %>%
        inner_join(models) %>%
        group_by(Training.file, Run, Pilot.data, Extrapolation, Weighting) %>%
        # https://stackoverflow.com/questions/32773617/add-column-of-predicted-values-to-data-frame-with-dplyr
        do(data.frame(., predicted.error=predict(.$model[[1]],.,type="response"))) %>%
        select(-model);
    return(preds);
}


#######################
#
# Plotting

# See https://stackoverflow.com/questions/14255533/pretty-ticks-for-log-normal-scale-using-ggplot2-dynamic-not-manual
# https://stackoverflow.com/questions/15178081/pretty-axis-labels-for-log-scale-in-ggplot
# https://stackoverflow.com/questions/11335836/increase-number-of-axis-ticks
# http://ggplot2.tidyverse.org/reference/annotation_logticks.html
#

residuals = predictions(eval_data, models %>% filter(Training.file %not_in% test_file_names)) %>%
    group_by(Training.file, Run, Pilot.data, Extrapolation, Weighting) %>%
    summarise(Residual=(predicted.error-min.av.error))

residual_plot = ggplot(data=residuals, mapping=aes(x=Weighting, y=Residual, fill=Extrapolation, group=interaction(Run,Extrapolation))) +
    geom_col(position="dodge") +
    facet_grid(Training.file ~ Pilot.data, scales="free") +
    theme_bw() + theme(axis.title.x=element_blank(), axis.title.y=element_blank()) +
    scale_fill_brewer(palette="PuBuGn", direction=-1)
# residual_plot

ggsave(filename='graphics/residuals.pdf', plot=residual_plot, width=9.0, height=5.5, units="in")


log10_breaks = function (n=10) {
    scales::trans_breaks("log10", function(x) 10^x, n=n)
}

log10_labels = function () {
    trans_format("log10", scales::math_format(10^.x))
}

# Only plot iVar models
#
pfile='ag_news'
prun='7'
pmodels = subset(models, Weighting=='n/e*(1-e)' & Run==prun & Training.file==pfile & Extrapolation=='a+b*n^c')
ex_plot = ggplot(data=train_data %>% filter(Run==prun & Training.file==pfile)) +
    # facet_wrap(facets='Training.file', ncol=4, scales='free') +
    geom_line(data=predictions(plot_data, pmodels %>% filter(Extrapolation=='a+b*n^c')), aes(training.examples, predicted.error, colour=Pilot.data, group=interaction(Run,Pilot.data)), alpha=0.7) +
    geom_point(aes(training.examples, error, colour=Pilot.data, shape=Pilot.data), size=2) +
    geom_point(data=subset(eval_data, Training.file==pfile), size=7, shape=8, colour='red', aes(training.examples, min.av.error)) +
    scale_x_log10(breaks=log10_breaks(2), labels=log10_labels(), name="") +
    scale_y_log10(breaks=scales::pretty_breaks(n=5), name="") +
    theme_bw() + theme(panel.grid.minor = element_blank(), panel.grid.major = element_blank()) +
    scale_colour_discrete(name="Pilot data") +
    scale_colour_brewer(palette="PuOr", direction=-1)
# ex_plot

ggsave(filename='graphics/extrapolate.pdf', plot=ex_plot, width=4.0, height=3.0, units="in")


#######################
#
# Tables

short_names = data.frame(Training.file=c('ag_news', 'amazon_review_full', 'amazon_review_polarity', 'dbpedia',
                                         'sogou_news', 'yahoo_answers', 'yelp_review_full', 'yelp_review_polarity'),
                         Short.name=c('AGN', 'ARF', 'ARP', 'DBP', 'SGN', 'YAA', 'YRF', 'YRP'))

shorten_names = function (d) {
    d %>% left_join(short_names) %>% mutate(Training.file=Short.name) %>% select(-Short.name)
}

# Evaluation table

edata = subset(eval_data, Training.file %in% test_file_names)
emodels = subset(models,
                 Extrapolation=='a+b*n^c'&Weighting=='n/e*(1-e)'&Training.file %in% test_file_names)
epreds = predictions(edata, emodels) %>%
    mutate(delta=predicted.error/min.av.error - 1) %>%
    ungroup %>%
    select(Training.file, Pilot.data, Run, delta)
etab1 = epreds %>%
    group_by(Training.file, Pilot.data) %>%
    # summarise(Mean=mean(abs(delta))) %>% 
    summarise(RMS=sqrt(mean(delta^2))) %>%
    ungroup %>%
    shorten_names %>%
    spread(key=Training.file, value=RMS)
etab2 = epreds %>%
    group_by(Pilot.data) %>%
    # summarise(Mean=mean(abs(delta)), RMS=sqrt(mean(delta^2)))
    summarise(Overall=sqrt(mean(delta^2)))

etab = left_join(etab1, etab2)

print.xtable(xtable(etab, auto=TRUE, digits=4), include.rownames=FALSE, booktabs=TRUE)

# Data statistics

ds = read.csv('file-stats.csv', header=T)

ds = ds %>%
    spread(key=type, value=training.examples) %>%
    mutate('Training file'=Training.file) %>%
    shorten_names %>%
    select('Training file',Training.file,nlabels,train,test)

print.xtable(xtable(ds, auto=TRUE), include.rownames=FALSE, booktabs=TRUE)

###############################################################
##
## Old stuff below here

## Multipanel extrapolation plot
## pmodels = subset(models, Weighting == 'n/e*(1-e)' & Training.file %not_in% test_file_names)
## ex_plot = ggplot(data=train_data %>% filter(Training.file %not_in% test_file_names)) +
##     # geom_jitter(data=raw_data %>% filter(Training.file %not_in% test_file_names), aes(training.examples, error), height=0, shape='.', alpha=I(0.07)) +
##     facet_wrap(facets='Training.file', ncol=4, scales='free') +
##     geom_line(data=(predictions(plot_data, pmodels %>% filter(Extrapolation=='a+b*n^-1/2')) %>% filter(predicted.error>=1e-2)), aes(training.examples, predicted.error, colour=Pilot.data, group=interaction(Run,Pilot.data)), alpha=0.7, linetype="dashed") +
##     geom_line(data=predictions(plot_data, pmodels %>% filter(Extrapolation=='b*n^c')), aes(training.examples, predicted.error, colour=Pilot.data, group=interaction(Run,Pilot.data)), alpha=0.7, linetype="dotted") +
##     geom_line(data=predictions(plot_data, pmodels %>% filter(Extrapolation=='a+b*n^c')), aes(training.examples, predicted.error, colour=Pilot.data, group=interaction(Run,Pilot.data)), alpha=0.7) +
##     geom_point(aes(training.examples, error, colour=Pilot.data, shape=Pilot.data), size=2) +
##     geom_point(data=subset(eval_data, Training.file %not_in% test_file_names), size=7, shape=8, colour='red', aes(training.examples, min.av.error)) +
##     scale_x_log10(breaks=log10_breaks(2), labels=log10_labels(), name="") +
##     scale_y_log10(breaks=scales::pretty_breaks(n=5), name="") +
##     theme_bw() + theme(panel.grid.minor = element_blank()) + # , panel.grid.major = element_blank()) +
##     scale_colour_discrete(name="Pilot data") +
##     # guides(shape=FALSE) +
##     scale_colour_brewer(palette="PuOr", direction=-1)
##
## ex_plot

## # https://stackoverflow.com/questions/44930704/r-nls-not-picking-up-additional-arguments-when-used-in-custom-function-in-geom

## weighted.lm = function(formula, data, ...) {
##   smooth.call = match.call()
##   # Change the function called
##   smooth.call[[1]] = eval(parse(text="lm"))
##   # add weights
##   data$error = data$y
##   data$training.examples = data$x
##   smooth.call$weights = weightfn(data)
##   eval.parent(smooth.call)
## }

## weighted.nls = function(formula, data, ...) {
##   smooth.call = match.call()
##   # Change the function called
##   smooth.call[[1]] = eval(parse(text="nls"))
##   # add weights
##   data$error = data$y
##   data$training.examples = data$x
##   smooth.call$weights = weightfn(data)
##   eval.parent(smooth.call)
## }


## # ggplot facets with facet_wrap(facets='Training.file', scales='free')
## # are the obvious way to build an array of plots.
## # However, ggplot2 doesn't currently support log-transformed
## # coordinates (with coord_trans) and scales='free' in faceting.
## # See https://github.com/tidyverse/ggplot2/issues/95
## #
## ## ggplot(data=merged) +
## ##     geom_jitter(data=data, aes(training.examples, error), height=0, shape='.', alpha=I(0.1)) +
## ##     # facet_wrap(facets='Training.file', scales='free') +
## ##     facet_wrap(facets='Training.file') +
## ##     # geom_smooth(aes(training.examples, error, colour=Configuration), method='weighted.lm', formula= y~I(1.0/sqrt(x)), linetype="dashed", fullrange=TRUE, se=FALSE) +
## ##     geom_smooth(aes(training.examples, error, colour=Configuration), method='weighted.nls', formula= y~plaw.selfStart(x,a,b,c), method.args=list(control=nls.control(warnOnly=TRUE,maxiter=100000,tol=1e-4,minFactor=1e-7)), fullrange=TRUE, se=FALSE) +
## ##     geom_point(aes(training.examples, error, colour=Configuration)) +
## ##     # scale_x_continuous(breaks=log_breaks(2)) +
## ##     # scale_y_continuous(breaks=log_breaks(7)) +
## ##     coord_trans(x="log10", y="log10") +
## ##     theme(panel.grid.minor = element_blank()) +
## ##     geom_point(data=target.100, size=7, colour='red', shape=8, aes(training.examples, error))

## #
## # So instead I am going for the more complex solution
## # See:
## # https://cran.r-project.org/web/packages/cowplot/vignettes/introduction.html
## # https://cran.r-project.org/web/packages/cowplot/vignettes/plot_grid.html

## log.jitter = function(xs, jitter=0.1) {  # Jitter appropriate for data that will be log-transformed
##     log10.1jitter = log10(1+jitter)
##     log10.jitters = runif(length(xs), -log10.1jitter, log10.1jitter)
##     10**(log10(xs)+log10.jitters)
## }

## # plot.training.file() generates a ggplot for a single training file
## #
## plot.training.file = function (tf, legend=FALSE) {
##     merged.tf = merged %>% filter(training.file==tf);
##     data.tf = data %>% filter(training.file==tf);
##     target.100.tf = target.100 %>% filter(training.file==tf);
##     plt = ggplot(data=merged.tf) +
##         geom_point(data=data.tf, aes(log.jitter(training.examples), error), shape='.', alpha=I(0.1)) +
##         geom_smooth(aes(training.examples, error, colour=Configuration),
##                     method='weighted.lm',
##                     formula= y~I(1.0/sqrt(x)),
##                     linetype="dashed", fullrange=TRUE, se=FALSE, n=1000) +
##         geom_smooth(aes(training.examples, error, colour=Configuration),
##                     method='weighted.nls',
##                     formula= y~plaw.selfStart(x,a,b,c),
##                     method.args=list(control=nls.control(warnOnly=TRUE,maxiter=100000,tol=1e-4,minFactor=1e-7)),
##                     fullrange=TRUE, se=FALSE, n=1000) +
##         geom_point(aes(training.examples, error, colour=Configuration)) +
##         scale_x_continuous(breaks=log_breaks(2), name="") +
##         scale_y_continuous(breaks=log_breaks(4), name="") +
##         coord_trans(x="log10", y="log10") +
##         geom_point(data=target.100.tf, size=7, colour='red', shape=8, aes(training.examples, error)) +
##         ggtitle(tf)
##     # Manipulating legends: http://www.cookbook-r.com/Graphs/Legends_(ggplot2)/
##     if (legend) {
##         plt = plt + scale_colour_discrete(name="Fitted on",
##                                           breaks=c('global.10', 'global.50', 'global.100', 'local.10', 'local.50', 'local.100'),
##                                           labels=c('==0.1', '==0.5', '==1.0', '<=0.1', '<=0.5', '<=1.0'))
##     }
##     else {
##         plt = plt + theme(legend.position="none")
##     }
##     plt
## }

## training.files = levels(merged$training.file)
## ntraining.files = length(training.files)

## # ps is a list of plots of all but the last training file, without legend
## ps = lapply(training.files[1:ntraining.files-1],
##             plot.training.file)

## # now add the plot of the last training file, with a legend
## ps = c(ps, list(plot.training.file(training.files[ntraining.files], legend=TRUE)))

## # mplt is a single plot combining all of the individual plots

## mplt = plot_grid(plotlist=ps, rel_widths=c(rep(1, times=ntraining.files-1), 1.3))

## save_plot(plot=mplt, filename='~/tmp/classifier-predictions.png', dpi=100, base_width=15, base_height=15)

## # ggsave(filename='classifier-predictions.png', dpi=100, height=7.0, width=7.0)
