#!/usr/bin/env python3

"""
Train and evaluate the fastText classifier on several different
sized fractions of several datasets with a range of different
hyperparameters, and produce a CSV file with the accuracies of
the run.

The command-line arguments should end in a list of training data 
file names, which should end in '.train'.  This program assumes
that there are corresponding test files ending in '.test'.

The training data is optionally randomly shuffled.

A fraction of the words in the training data, and the labels in
the training data, can be replaced with randomly generated words or
labels, drawn from the training data's marginal distributions.

Then a subset of the training data is selected for training the
classifier.

This is done in such a way that subset relationships hold between
the smaller and larger training datasets, or the sets of random
permutations made. 
"""

import ft  # fasttext package

import argparse
import csv
import gc
import multiprocessing
import pathlib
import random
import sys


def unlink(path):
  """Unlinks a pathlib path if it exists"""
  if path.exists():
    path.unlink()


def parse_args():
  """Parse command-line arguments"""
  def list_of_vals(t):
    """Recognises a list of comma-separated values of type t"""
    return lambda s: [t(v) for v in s.split(',')]
  p = argparse.ArgumentParser(description=__doc__,
                              formatter_class=argparse.RawDescriptionHelpFormatter)
  p.add_argument('--tmp-dir', '-t', default="./tmp",
                 help="temporary files directory (will be erased)")
  p.add_argument('--fractions', '-f', type=list_of_vals(float),
                 default=[0.01,0.02,0.05,0.1,0.2,0.5,1.0],
                 # default=[0.01,0.015,0.02,0.03,0.05,0.07,0.1,0.15,0.2,0.3,0.5,0.7,1.0],
                 help='fractions of training data')
  p.add_argument('--lr', '-l', type=list_of_vals(float),
                 default=[2.0],
                 # default=[0.01,0.02,0.05,0.1,0.2,0.5,1.0,2.0,5.0,10.0],
                 help='learning rates')
  p.add_argument('--dim', '-d', type=list_of_vals(int),
                 default=[20],
                 # default=[1,2,3,5,7,10,20,50,100],
                 help='size of word vectors')
  p.add_argument('--loss', '-L', type=list_of_vals(str),
                 default=['ns'],
                 # default=['ns','hs','softmax'],
                 help='loss function')
  p.add_argument('--wordNgrams', '-w', type=list_of_vals(int),
                 default=[3],
                 # default=[1,2,3,4],
                 help='ngram length')
  p.add_argument('--data-noise', '-X', type=list_of_vals(float),
                 default=[0.0,0.01,0.02,0.05,0.1,0.2,0.5,0.7,0.8,0.9,0.95,0.975,0.99,0.995,0.9975,0.999,1.0],
                 # default=[0.0,0.01,0.02,0.05,0.1,0.2,0.5,1.0],
                 help='fraction of data items replaced with random noise')
  p.add_argument('--label-noise', '-Y', type=list_of_vals(float),
                 default=[0.0,0.01,0.02,0.05,0.1,0.2,0.5,1.0],
                 help='fraction of labels replaced with random noise')
  p.add_argument('--shuffle', '-s', action='store_true',
                 help="shuffle training data before choosing each subset")
  p.add_argument('--processes', '-p', type=int, default=32,
                 help='number of simultaneous fastText processes')
  p.add_argument('--threads', '-T', type=int, default=4,
                 help='number of threads per fastText process')
  p.add_argument('--output', '-o', type=argparse.FileType('w'), default=sys.stdout,
                 help='CSV results file')
  p.add_argument('training_files', nargs='+', 
                 help="training files (should end in '.train')")
  return p.parse_args()


csvheader = ['training_file','data_noise','label_noise','fraction','size','config','lr','dim','loss','wordNgrams','prec','rec']


def add_label_noise(source, shuffled_labelpos, shuffled_labels, noisefrac):
  """Permutes the labels in noise fraction of source"""
  if noisefrac > 0:
    source = list(source)  # top-level copy
    for i in range(int(noisefrac*len(source) + 0.5)):
      j = shuffled_labelpos[i]
      source_j = source[j].split(maxsplit=1)
      source[j] = shuffled_labels[i]+' '+source_j[1]
  return source


def add_data_noise(source, shuffled_datapos, shuffled_data, noisefrac):
  """Permutes the words in noisefrac of source"""
  if noisefrac > 0:
    source = list(source)  # top-level copy
    m = int(noisefrac*len(shuffled_datapos) + 0.5)
    ss = sorted(zip(shuffled_datapos[:m], shuffled_data[:m]), reverse=True)
    n = 0
    for i in range(len(source)):
      changed_flag = False
      line = source[i].split()
      len_line1 = len(line)-1
      while ss and ss[-1][0]-n < len_line1:
        assert ss[-1][0] >= n
        line[ss[-1][0]-n+1] = ss[-1][1]
        changed_flag = True
        ss.pop()
      if changed_flag:
        source[i] = ' '.join(line)
      n += len_line1
    assert not ss
    assert n == len(shuffled_data)
  return source

  
def copy_frac(source, frac, outputpath):
  """Copy fraction frac from source to outputpath file"""
  nlines = int(frac*len(source) + 0.5)
  with outputpath.open('w', encoding='utf-8') as outf:
    for line in source[:nlines]:
      print(line, file=outf)
  return nlines


def train_test1(args):
  """Trains a classifier on a fractional training file with a specified config,
  and tests it on the test file.

  Returns a tuple: (id,lr,dim,loss,wordNgrams,prec,rec)"""
  (id,config),fractional_train_file,test_file,tmpdir,threads = args # Unpack args
  model_path = pathlib.Path(tmpdir)/('model_'+str(id))
  model_file = str(model_path)
  precision = 'NA'
  recall = 'NA'
  if ft.supervised(fractional_train_file, model_file, thread=threads, **config):
    result = ft.test(model_file, test_file)
    if result:
      precision = result.precision
      recall = result.recall
  unlink(model_path.with_suffix('.bin'))
  unlink(model_path.with_suffix('.vec'))
  
  return (id,config['lr'],config['dim'],config['loss'],config['wordNgrams'],precision,recall)

    
def train_test(args, train_file, train, data_noise, label_noise, frac, configs, outf):
  """For each of the configurations specified in configs, train on fraction
  frac of train_file, then test on corresponding test file, and write
  results to outf.
  """
  train_file_stem = train_file[:-6] if train_file.endswith('.train') else train_file
  test_file = train_file_stem+'.test'
  test_path = pathlib.Path(test_file)
  assert(test_path.exists())
  train_stem = train_file_stem.split('/')[-1] # ignore all but the last component

  threads = args.threads
  tmpdir = pathlib.Path(args.tmp_dir)
  tmpdir.mkdir(parents=True, exist_ok=True)
  fractional_train_path = tmpdir/'train'
  nlines = copy_frac(train, frac, fractional_train_path)
  fractional_train_file = str(fractional_train_path)

  nconfigs = len(configs)
  if nconfigs > 1:
    with multiprocessing.Pool(processes=min(args.processes, nconfigs), 
                              maxtasksperchild=1) as pool:
      full_configs = list((config, fractional_train_file, test_file, str(tmpdir), threads) 
                          for config in configs)
      random.shuffle(full_configs)
      for result in pool.imap_unordered(train_test1, full_configs, chunksize=1):
        outf.writerow((train_stem,data_noise,label_noise,frac,nlines)+result)
        args.output.flush()
      pool.close()
      pool.join()
  else:
    full_config = (configs[0], fractional_train_file, test_file, str(tmpdir), threads)
    result = train_test1(full_config)
    outf.writerow((train_stem,data_noise,label_noise,frac,nlines)+result)
    args.output.flush()

  fractional_train_path.unlink()  # delete fractional training file
    

if __name__ == '__main__':
  args = parse_args()

  configs = ({'lr':lr, 'dim':dim, 'loss':loss, 'wordNgrams':wordNgrams}
             for lr in args.lr 
             for dim in args.dim
             for loss in args.loss
             for wordNgrams in args.wordNgrams
           )
  configs = list(enumerate(configs))

  outf = csv.writer(args.output)
  outf.writerow(csvheader)
  args.output.flush()

  for train_file in args.training_files:
    train = [line.strip() for line in open(train_file, mode="rU", encoding='utf-8-sig')]

    if args.shuffle:
      random.shuffle(train)

    shuffled_labels = [line.split(maxsplit=1)[0] for line in train]
    random.shuffle(shuffled_labels)
    ntrain = len(shuffled_labels)
    shuffled_labelpos = random.sample(range(ntrain), k=ntrain)

    shuffled_data = [word for line in train for word in line.split()[1:]]
    random.shuffle(shuffled_data)
    ndata = len(shuffled_data)
    shuffled_datapos = random.sample(range(ndata), k=ndata)

    for data_noise in args.data_noise:
      data_noisified_train = add_data_noise(train, shuffled_datapos, shuffled_data, data_noise)
      for label_noise in args.label_noise:
        label_data_noisified_train = add_label_noise(data_noisified_train, shuffled_labelpos, shuffled_labels, label_noise)
        gc.collect()
        for frac in args.fractions:
          train_test(args, train_file, label_data_noisified_train, data_noise, label_noise, frac, configs, outf)
