#!/usr/bin/env python3

"""Python interface to fasttext package.

Mark Johnson, 7th January 2018, modified 19th February 2018

Calls fasttext as a subprocess, and collects output.
"""

import collections
import re
import subprocess
import sys

def call(mode, program="fasttext", **args):
  """Construct a subprocess call to fasttext mode.  Passes all keyword args
  to fasttext.  Returns the subprocess.CompletedProcess object for the process."""
  subproc_args = [program, mode]
  for key, val in args.items():
    subproc_args.extend(('-'+key, str(val)))
  subproc = subprocess.run(subproc_args,
                           stdin=subprocess.DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                           universal_newlines=True)
  return subproc


def supervised(train_file, model_file, verbose=0, **args):
  """Calls `fasttext supervised` to train a supervised classifier.  All
  args are passed to fasttext.  
  Returns True if the training process succeed, and False if it failed."""
  p = call('supervised', input=train_file, output=model_file, verbose=verbose, **args)
  # exit(p)
  return p.returncode == 0


testResults = collections.namedtuple('TestResults', 'N precision recall')

test_rex=re.compile(r"""N\t(?P<N>[0-9]+)\n
                        P@1\t(?P<P>[+-]?[0-9]+[.]?[0-9]*)\n
                        R@1\t(?P<R>[+-]?[0-9]+[.]?[0-9]*)""",
                        # ^Number of examples:\t+(?P<Nexamples>[0-9]+)\s*$""", 
                    re.VERBOSE)

def test(model_file, test_file, k=1, program="fasttext"):
  """Calls `fasttext test` to test a supervised classifier.  Parses the
  output string to return precision, recall, etc."""
  subproc = subprocess.run([program,'test',model_file+'.bin',test_file],
                           stdin=subprocess.DEVNULL, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                           universal_newlines=True)
  if subproc.returncode:
    return None # fastText test subprocess crashed (why?)
  mo = test_rex.match(subproc.stdout)
  if not mo:
    sys.exit('unexpected output from fasttext test:\n{}'.format(subproc.stdout.__repr__()))
  return testResults(N=int(mo.group('N')), 
                     precision=float(mo.group('P')), 
                     recall=float(mo.group('R')))
                     # Nexamples=int(mo.group('Nexamples')))
  
